@@ -75,6 +75,70 @@ class PyProxyBaseTestCase(unittest.TestCase):
         proxy = MyProxy3('notused')
         self.assertEqual(list(proxy), list('another'))
 
+    def test_string_to_int(self):
+        # Strings don't have the tp_number.tp_int pointer
+        proxy = self._makeOne("14")
+        self.assertEqual(14, int(proxy))
+
+    def test_custom_int_to_int(self):
+        class CustomClass(object):
+            def __int__(self):
+                return 42
+        proxy = self._makeOne(CustomClass())
+        self.assertEqual(42, int(proxy))
+
+    def test_string_to_float(self):
+        proxy = self._makeOne("14")
+        self.assertEqual(float("14"), float(proxy))
+
+    def test_incorrect_string_to_int(self):
+        proxy = self._makeOne("")
+        self.assertRaises(ValueError, int, proxy)
+
+    def test_incorrect_string_to_float(self):
+        proxy = self._makeOne("")
+        self.assertRaises(ValueError, float, proxy)
+
+    def test_custom_float_to_float(self):
+        class CustomClass(object):
+            def __float__(self):
+                return 42.0
+        proxy = self._makeOne(CustomClass())
+        self.assertEqual(42.0, float(proxy))
+
+    def test___unicode__of_unicode(self):
+        from zope.proxy._compat import PY3, _u
+        if PY3: # Gone in Python 3:
+            return
+        s = _u('Hello, \u2603')
+        proxy = self._makeOne(s)
+        self.assertEqual(unicode(proxy), s)
+
+    def test___unicode__of_custom_class(self):
+        from zope.proxy._compat import PY3, _u
+        if PY3: # Gone in Python 3:
+            return
+        class CustomClass(object):
+            def __unicode__(self):
+                return _u('Hello, \u2603')
+        cc = CustomClass()
+        self.assertEqual(unicode(cc), _u('Hello, \u2603'))
+        proxy = self._makeOne(cc)
+        self.assertEqual(unicode(proxy), _u('Hello, \u2603'))
+
+    def test___unicode__of_custom_class_no_unicode(self):
+        # The default behaviour should be preserved
+        from zope.proxy._compat import PY3, _u
+        if PY3: # Gone in Python 3:
+            return
+        class CustomClass(object):
+            pass
+        cc = CustomClass()
+        cc_unicode = unicode(cc)
+        self.assertEqual(type(cc_unicode), unicode)
+        proxy = self._makeOne(cc)
+        self.assertEqual(unicode(proxy), cc_unicode)
+
     def test___call__(self):
         def _foo():
             return 'FOO'
@@ -418,7 +482,7 @@ class PyProxyBaseTestCase(unittest.TestCase):
             "complex(x)",
         ]
         if not PY3: # long is gone in Python 3
-            ops.append("long(x)") 
+            ops.append("long(x)")
         return ops
 
     def test_unops(self):
@@ -581,6 +645,150 @@ class PyProxyBaseTestCase(unittest.TestCase):
         w = self._makeOne(o)
         self.assertTrue(w.__class__ is o.__class__)
 
+    def test_descriptor__set___only_in_proxy_subclass(self):
+
+        class Descriptor(object):
+            value = None
+            instance = None
+            def __set__(self, instance, value):
+                self.value = value
+                self.instance = instance
+
+        descriptor = Descriptor()
+        class Proxy(self._getTargetClass()):
+            attr = descriptor
+
+        proxy = Proxy(object())
+        proxy.attr = 42
+
+        self.assertEqual(proxy.attr, descriptor)
+        self.assertEqual(descriptor.value, 42)
+        self.assertEqual(descriptor.instance, proxy)
+
+    def test_descriptor__get___set___in_proxy_subclass(self):
+
+        class Descriptor(object):
+            value = None
+            instance = None
+            cls = None
+
+            def __get__(self, instance, cls):
+                self.cls = cls
+                return self.value
+
+            def __set__(self, instance, value):
+                self.value = value
+                self.instance = instance
+
+        descriptor = Descriptor()
+        descriptor.value = "descriptor value"
+        class Proxy(self._getTargetClass()):
+            attr = descriptor
+
+        proxy = Proxy(object())
+        self.assertEqual(proxy.attr, "descriptor value")
+        self.assertEqual(descriptor.cls, Proxy)
+
+        proxy.attr = 42
+
+        self.assertEqual(descriptor.value, 42)
+        self.assertEqual(descriptor.instance, proxy)
+
+    def test_non_descriptor_in_proxy_subclass__dict__(self):
+        # Non-descriptors in the class dict of the subclass
+        # are always passed through to the wrapped instance
+        class Proxy(self._getTargetClass()):
+            attr = "constant value"
+
+        proxy = Proxy(object())
+        self.assertEqual(proxy.attr, "constant value")
+
+        self.assertRaises(AttributeError, setattr, proxy, 'attr', 42)
+        self.assertEqual(proxy.attr, "constant value")
+
+    def _check_wrapping_builtin_returns_correct_provided_by(self, proxy_class, builtin_type):
+        # We get the __implemented__ (fallback) of the type, not our own
+        from zope.interface import Interface
+        from zope.interface import classImplements
+        from zope.interface import classImplementsOnly
+        from zope.interface import implementedBy
+        from zope.interface import providedBy
+        from zope.interface import implementedBy
+
+        # Set up the builtin interface
+        class IFoo(Interface):
+            pass
+        impl_before = list(implementedBy(builtin_type))
+
+        classImplements(builtin_type, IFoo)
+
+        builtin = builtin_type()
+        self.assertTrue(IFoo in list(providedBy(builtin)))
+        self.assertTrue(IFoo in list(implementedBy(builtin_type)))
+
+        try:
+            # The asserts must be before we remove the interface
+            # because there's a single object that gets mutated
+
+            proxy_instance = proxy_class(builtin)
+            provided_instance = providedBy(proxy_instance)
+            self.assertTrue(IFoo in list(provided_instance))
+
+            # XXX: PyPy 2.5.0 has a bug where proxys around types
+            # aren't correctly hashable, which breaks this part of the
+            # test. This is fixed in 2.5.1, but as of 2015-05-28,
+            # TravisCI still uses 2.5.0.
+            proxy_type = proxy_class(builtin_type)
+            from zope.interface.declarations import BuiltinImplementationSpecifications
+            if proxy_type in BuiltinImplementationSpecifications \
+               and BuiltinImplementationSpecifications.get(proxy_type, self) is not self:
+                provided_type = implementedBy(proxy_type)
+                self.assertTrue(IFoo in list(provided_type))
+            else:
+                import sys
+                self.assertEqual((2,5,0), sys.pypy_version_info[:3])
+        finally:
+            classImplementsOnly(builtin_type, *impl_before)
+
+    def test_wrapping_builtin_type_returns_correct_provided_by(self):
+        self._check_wrapping_builtin_returns_correct_provided_by(self._getTargetClass(), list)
+
+    def _check_wrapping_builtin_with_subclass_returns_correct_provided_by(self, builtin_type):
+        class Proxy(self._getTargetClass()):
+            pass
+
+        self._check_wrapping_builtin_returns_correct_provided_by(Proxy, builtin_type)
+        # Our new class did not gain an __implemented__ attribute, unless we're
+        # the pure-python version
+        if hasattr(Proxy, '__implemented__'):
+            from zope.proxy import PyProxyBase
+            self.assertTrue(self._getTargetClass() is PyProxyBase)
+
+    def test_wrapping_builtin_with_subclass_returns_correct_provided_by(self):
+        self._check_wrapping_builtin_with_subclass_returns_correct_provided_by(list)
+
+    def test_method_in_proxy_subclass(self):
+        class Proxy(self._getTargetClass()):
+            def __getitem__(self, k):
+                return k
+
+        proxy = Proxy(object())
+        # Both when called by the interpreter, which bypasses
+        # __getattribute__
+        self.assertEqual(proxy[42], 42)
+        # And when asked for as an attribute
+        self.assertNotEqual(getattr(proxy, '__getitem__'), self)
+
+    def test_string_to_int(self):
+        # XXX Implementation difference: This works in the
+        # Pure-Python version, but fails in CPython.
+        # See https://github.com/zopefoundation/zope.proxy/issues/4
+        proxy = self._makeOne("14")
+        try:
+            self.assertEqual(14, int(proxy))
+        except TypeError:
+            from zope.proxy import PyProxyBase
+            self.assertNotEqual(self._getTargetClass(), PyProxyBase)
 
 class ProxyBaseTestCase(PyProxyBaseTestCase):
 
@@ -588,6 +796,88 @@ class ProxyBaseTestCase(PyProxyBaseTestCase):
         from zope.proxy import ProxyBase
         return ProxyBase
 
+class Test_py__module(unittest.TestCase):
+    # Historically, proxying __module__ has been troublesome,
+    # especially when subclasses of the proxy class are involved;
+    # there was also a discrepancy between the C and Python implementations
+    # in that the C implementation only failed Test_subclass__module:test__module__in_instance,
+    # whereas the Python version failed every test.
+    # See https://github.com/zopefoundation/zopetoolkit/pull/2#issuecomment-106075153
+    # and https://github.com/zopefoundation/zope.proxy/pull/8
+
+    def _getTargetClass(self):
+        from zope.proxy import PyProxyBase
+        return PyProxyBase
+
+    def _makeProxy(self, obj):
+        from zope.proxy import PyProxyBase
+        return self._getTargetClass()(obj)
+
+    def _check_module(self, obj, expected):
+        self.assertEqual(expected, obj.__module__)
+        self.assertEqual(expected, self._makeProxy(obj).__module__)
+
+    def test__module__in_instance(self):
+        # We can find __module__ in an instance dict
+        class Module(object):
+            def __init__(self):
+                self.__module__ = 'module'
+
+        self._check_module(Module(), 'module')
+
+    def test__module__in_class_instance(self):
+        # We can find module in an instance of a class
+        class Module(object):
+            pass
+
+        self._check_module(Module(), __name__)
+
+    def test__module__in_class(self):
+        # We can find module in a class itself
+        class Module(object):
+            pass
+        self._check_module(Module, __name__)
+
+    def test__module_in_eq_transitive(self):
+        # An object that uses __module__ in its implementation
+        # of __eq__ is transitively equal to a proxy of itself.
+        # Seen with zope.interface.interface.Interface
+
+        class Module(object):
+            def __init__(self):
+                self.__module__ = __name__
+            def __eq__(self, other):
+                return self.__module__ == other.__module__
+
+        module = Module()
+        # Sanity checks
+        self.assertEqual(module, module)
+        self.assertEqual(module.__module__, __name__)
+
+        # transitive equal
+        self.assertEqual(module, self._makeProxy(module))
+        self.assertEqual(self._makeProxy(module), module)
+
+class Test__module(Test_py__module):
+
+    def _getTargetClass(self):
+        from zope.proxy import ProxyBase
+        return ProxyBase
+
+class Test_py_subclass__module(Test_py__module):
+
+    def _getTargetClass(self):
+        class ProxySubclass(super(Test_py_subclass__module,self)._getTargetClass()):
+            pass
+        return ProxySubclass
+
+class Test_subclass__module(Test__module):
+
+    def _getTargetClass(self):
+        class ProxySubclass(super(Test_subclass__module,self)._getTargetClass()):
+            pass
+        return ProxySubclass
+
 
 class Test_py_getProxiedObject(unittest.TestCase):
 
@@ -620,7 +910,6 @@ class Test_py_getProxiedObject(unittest.TestCase):
         proxy2 = self._makeProxy(proxy)
         self.assertTrue(self._callFUT(proxy2) is proxy)
 
-
 class Test_getProxiedObject(Test_py_getProxiedObject):
 
     def _callFUT(self, *args):
@@ -1124,4 +1413,8 @@ def test_suite():
         unittest.makeSuite(Test_removeAllProxies),
         unittest.makeSuite(Test_ProxyIterator),
         unittest.makeSuite(Test_nonOverridable),
+        unittest.makeSuite(Test_py__module),
+        unittest.makeSuite(Test__module),
+        unittest.makeSuite(Test_py_subclass__module),
+        unittest.makeSuite(Test_subclass__module),
     ))
